import logging
import sys


class Formatter(logging.Formatter):
    def format(self, record):
        message = record.msg % record.args if record.args else record.msg
        if record.levelno == logging.INFO:
            return message
        else:
            return f"{record.levelname} {message}"


def init():
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(Formatter())
    logging.root.addHandler(handler)
    logging.root.setLevel(logging.INFO)


def show_debug():
    logging.root.setLevel(logging.DEBUG)



import dataclasses
import json
import logging
import sys
from decimal import Decimal

import requests

from pricehist.price import Price

from .basesource import BaseSource


class CoinDesk(BaseSource):
    def id(self):
        return "coindesk"

    def name(self):
        return "CoinDesk Bitcoin Price Index"

    def description(self):
        return (
            "An average of Bitcoin prices across leading global exchanges. \n"
            "Powered by CoinDesk, https://www.coindesk.com/price/bitcoin"
        )

    def source_url(self):
        return "https://www.coindesk.com/coindesk-api"

    def start(self):
        return "2010-07-17"

    def types(self):
        return ["close"]

    def notes(self):
        return ""

    def symbols(self):
        url = "https://api.coindesk.com/v1/bpi/supported-currencies.json"
        response = self.log_curl(requests.get(url))
        # TODO handle non-200
        # TODO handle can't parse JSON
        data = json.loads(response.content)
        # TODO handle no data
        relevant = [i for i in data if i["currency"] not in ["XBT", "BTC"]]
        return [
            (f"BTC/{i['currency']}", f"Bitcoin against {i['country']}")
            for i in sorted(relevant, key=lambda i: i["currency"])
        ]

    def fetch(self, series):
        if series.base != "BTC" or series.quote == "BTC":
            # BTC is the only valid base. BTC as the quote will return BTC/USD.
            logging.critical(
                f"Invalid pair '{'/'.join([series.base, series.quote])}'. "
                f"Run 'pricehist source {self.id()} --symbols' to list valid pairs."
            )
            sys.exit(1)
        data = self._data(series)
        prices = []
        for (d, v) in data["bpi"].items():
            prices.append(Price(d, Decimal(str(v))))
        return dataclasses.replace(series, prices=prices)

    def _data(self, series):
        url = "https://api.coindesk.com/v1/bpi/historical/close.json"
        params = {
            "currency": series.quote,
            "start": series.start,
            "end": series.end,
        }
        response = self.log_curl(requests.get(url, params=params))
        # TODO handle 404: can be 1) quote not found 2) date range invalid 3) no data(?)
        # TODO handle non-200 response (500 can be due to quote=XBT)
        # TODO handle can't parse JSON
        return json.loads(response.content)



import dataclasses
import logging
import sys
from datetime import datetime, timedelta
from decimal import Decimal

import requests
from lxml import etree

from pricehist import isocurrencies
from pricehist.price import Price

from .basesource import BaseSource


class ECB(BaseSource):
    def id(self):
        return "ecb"

    def name(self):
        return "European Central Bank"

    def description(self):
        return "European Central Bank Euro foreign exchange reference rates"

    def source_url(self):
        return "https://www.ecb.europa.eu/stats/exchange/eurofxref/html/index.en.html"

    def start(self):
        return "1999-01-04"

    def types(self):
        return ["reference"]

    def notes(self):
        return ""

    def symbols(self):
        root = self._data(more_than_90_days=True)
        nodes = root.cssselect("[currency]")
        currencies = sorted(set([n.attrib["currency"] for n in nodes]))
        # TODO handle no data
        iso = isocurrencies.by_code()
        return [(f"EUR/{c}", f"Euro against {iso[c].name}") for c in currencies]

    def fetch(self, series):
        if series.base != "EUR":  # EUR is the only valid base.
            logging.critical(
                f"Invalid pair '{'/'.join([series.base, series.quote])}'. "
                f"Run 'pricehist source {self.id()} --symbols' to list valid pairs."
            )
            sys.exit(1)
        almost_90_days_ago = (datetime.now().date() - timedelta(days=85)).isoformat()
        root = self._data(series.start < almost_90_days_ago)

        all_rows = []
        for day in root.cssselect("[time]"):
            date = day.attrib["time"]
            for row in day.cssselect(f"[currency='{series.quote}']"):
                rate = Decimal(row.attrib["rate"])
                all_rows.insert(0, (date, rate))
        # TODO if (no data && known symbol) { fail }
        selected = [
            Price(d, r) for d, r in all_rows if d >= series.start and d <= series.end
        ]

        return dataclasses.replace(series, prices=selected)

    def _data(self, more_than_90_days=False):
        url_base = "https://www.ecb.europa.eu/stats/eurofxref"
        if more_than_90_days:
            source_url = f"{url_base}/eurofxref-hist.xml"  # since 1999
        else:
            source_url = f"{url_base}/eurofxref-hist-90d.xml"  # last 90 days

        response = self.log_curl(requests.get(source_url))
        # TODO handle response not 200
        root = etree.fromstring(response.content)
        # TODO handle XML can't be parsed
        return root



import dataclasses
import json
from datetime import datetime
from decimal import Decimal

import requests

from pricehist.price import Price

from .basesource import BaseSource


class CoinMarketCap(BaseSource):
    def id(self):
        return "coinmarketcap"

    def name(self):
        return "CoinMarketCap"

    def description(self):
        return "The world's most-referenced price-tracking website for cryptoassets"

    def source_url(self):
        return "https://coinmarketcap.com/"

    def start(self):
        return "2013-04-28"

    def types(self):
        return ["mid", "open", "high", "low", "close"]

    def notes(self):
        return (
            "This source makes unoffical use of endpoints that power CoinMarketCap's "
            "public web interface. The price data comes from a public equivalent of "
            "the OHLCV Historical endpoint found in CoinMarketCap's official API.\n"
            "CoinMarketCap currency symbols are not necessarily unique, so it "
            "is recommended that you use IDs, which can be listed via the "
            "--symbols option. For example, 'ETH/BTC' is 'id=1027/id=1'. The "
            "corresponding symbols will be used in output."
        )

    def symbols(self):
        data = self._symbol_data()
        ids = [f"id={i['id']}" for i in data]
        descriptions = [f"{i['symbol'] or i['code']} {i['name']}".strip() for i in data]
        return list(zip(ids, descriptions))

    def fetch(self, series):
        data = self._data(series)

        prices = []
        # TODO handle when data has no "quotes" key (e.g. asking for EUR/USD)
        for item in data["data"]["quotes"]:
            d = item["time_open"][0:10]
            amount = self._amount(next(iter(item["quote"].values())), series.type)
            prices.append(Price(d, amount))

        output_base, output_quote = self._output_pair(series.base, series.quote, data)

        return dataclasses.replace(
            series, base=output_base, quote=output_quote, prices=prices
        )

    def _data(self, series):
        url = "https://web-api.coinmarketcap.com/v1/cryptocurrency/ohlcv/historical"

        params = {}

        if series.base.startswith("ID="):
            params["id"] = series.base[3:]
        else:
            params["symbol"] = series.base

        if series.quote.startswith("ID="):
            params["convert_id"] = series.quote[3:]
        else:
            params["convert"] = series.quote

        params["time_start"] = int(
            int(datetime.strptime(series.start, "%Y-%m-%d").timestamp())
        )
        params["time_end"] = (
            int(datetime.strptime(series.end, "%Y-%m-%d").timestamp()) + 24 * 60 * 60
        )  # round up to include the last day

        response = self.log_curl(requests.get(url, params=params))

        return json.loads(response.content)

    def _amount(self, data, type):
        if type in ["mid"]:
            high = Decimal(str(data["high"]))
            low = Decimal(str(data["low"]))
            return sum([high, low]) / 2
        else:
            return Decimal(str(data[type]))

    def _output_pair(self, base, quote, data):
        data_base = data["data"]["symbol"]
        data_quote = next(iter(data["data"]["quotes"][0]["quote"].keys()))

        lookup_quote = False
        if quote.startswith("ID="):
            symbols = {i["id"]: (i["symbol"] or i["code"]) for i in self._symbol_data()}
            lookup_quote = symbols[int(quote[3:])]

        output_base = data_base
        output_quote = lookup_quote or data_quote

        return (output_base, output_quote)

    def _symbol_data(self):
        fiat_url = "https://web-api.coinmarketcap.com/v1/fiat/map?include_metals=true"
        fiat_res = self.log_curl(requests.get(fiat_url))
        fiat = json.loads(fiat_res.content)
        crypto_url = (
            "https://web-api.coinmarketcap.com/v1/cryptocurrency/map?sort=cmc_rank"
        )
        crypto_res = self.log_curl(requests.get(crypto_url))
        crypto = json.loads(crypto_res.content)
        return crypto["data"] + fiat["data"]
